Unit IODrives;

(*
  Version 1.00 - Released January 19 1984.

  This unit was written after a long research.
  It's completely freeware. If you're going to use it in 
  a professional or a commercial software feel free to donate 
  $5 U.S. currency to the address below. 
  This unit was written under BP 7.0 and it makes use of some 
  new calls (e.g. Continue) that were not available in previous 
  Turbo Pascal versions so you'll need some fine tuning before
  it will work with previous versions.
  It can be compiled on all three platforms: DOS, DPMI and Windows.
  There's no need to put any special definitions. 
  Try the TESTDRV.PAS program which I wrote on all three platforms 
  and you'll find that it works just fine. 
  There are a drives which I can't yet recognize like XTraDrive 
  and SuperStor and if anyone knows how  to find these drives 
  I'll be glad to add them to this source.

  I can be reached on the on:
    Schwartz Gabriel
    EliaShim microcomputers, Inc.
    Tel: (972)-4-516111
    Fax: (972)-4-528613
    BBS: (972)-4-516113
    CompuServe id:72620,701
    Internet address:gabi@eliashim.co.il

  My address is:
    Schwartz Gabriel
    Hakanaim 33/3
    Arad 80700
    ISRAEL
*)

{$C MOVEABLE PRELOAD PERMANENT}

Interface

Type
  TDriveMedia = (INVALID_DRIVE, FLOPPY_DRIVE, REMOVABLE_DRIVE, FIXED_DRIVE,
                 LAN_DRIVE, RAM_DRIVE, CDROM_DRIVE, BERNOULLI_DRIVE,
                 SUBSTED_DRIVE, ASSIGNED_DRIVE,
                 DOUBLEDISK_DRIVE, DOUBLESPACE_DRIVE, STACKER_DRIVE,
                 TAPE_DRIVE, OPTICAL_DRIVE);

  PDrive = ^TDrive;
  TDrive = record
             Media : TDriveMedia; (* Drive media type                                   *)
             Path  : String;      (* Original Directory for substed and assigned drives *)
           end;

  PDrives = ^TDrives;
  TDrives = Array [1..26] of TDrive;

Var
  Drives : PDrives;

{------------------------------------------------------------------------------}
procedure FindDrives (var Drives : PDrives);
{------------------------------------------------------------------------------}
function  Media2String (Media : TDriveMedia) : String;
{------------------------------------------------------------------------------}

Implementation

Uses
{$IFDEF MSDOS}
    DOS
{$ELSE}
    WinDOS,
  {$IFDEF Windows}
     WinProcs, WinTypes
  {$ELSE}
     WinAPI
  {$ENDIF}
  {$ENDIF}
     , Strings, DPMI;

{$IFNDEF WINDOWS}
Const
  DRIVE_INVALID   = 0;
  DRIVE_REMOVABLE = 1;
  DRIVE_FIXED     = 2;
  DRIVE_REMOTE    = 3;
{$ENDIF}

Type
  TDrivesArray = Array [1..26] of Byte;

  PDPB = ^TDPB;
  TDPB = record
           Drive                : Byte;     (* 00h = A:, 01h = B:, etc                                      *)
           UnitNumber           : Byte;     (* Unit number within device driver                             *)
           BytesPerSector       : Word;     (* Bytes per sector                                             *)
           HighestSector        : Byte;     (* Highest sector number within a cluster                       *)
           ShiftCount           : Byte;     (* Shift count to convert clusters into sectors                 *)
           ReservedSectors      : Word;     (* Number of reserved sectors at beginning of drive             *)
           FAT                  : Byte;     (* Number of FATs                                               *)
           RootEntries          : Word;     (* Number of root directory entries                             *)
           FirstDataSector      : Word;     (* Number of first sector containing user data                  *)
           HighestCluster       : Word;     (* Highest cluster number (number of data clsters + 1)
                                               16-bit FAT if greater than 0FF6h, else 12-bit FAT            *)
           SectorsPerFAT        : Word;     (* Number of sectors per FAT                                    *)
           FirstDirectorySector : Word;     (* Sector number of first directory sector                      *)
           DeviceDriverHeader   : Pointer;  (* Address of device driver header                              *)
           MediaID              : Byte;     (* Media ID Byte                                                *)
           DiskAccessed         : Byte;     (* 00h if disk accessed, FFh if not                             *)
           NextDPB              : PDPB;     (* Pointer to next DPB                                          *)
           FreeCluster          : Word;     (* Cluster at which to start search for free space when writing,
                                               usually of free clusters allocated                           *)
           FreeClustersNumber   : Word;     (* Number of free clusters on drive, FFFFh = unknown            *)
           FAT16                : Boolean;  (* FAT is 16bit                                                 *)
           EntriesPerSector     : Byte;     (* BytesPerSector div 32                                        *)
           MaxSectors           : Word;     (* RootEntries div EntriesPerSector (ROOT only!)                *)
         end;

  P1stDPB = ^T1stDPB;      (* This record is only for translation to the above record *)
  T1stDPB = record
              Case Integer of
                    (* For DOS 3.xx *)
                0 : (_Drive                : Byte;    (* 00h = A:, 01h = B:, etc                                      *)
                     _UnitNumber           : Byte;    (* Unit number within device driver                             *)
                     _BytesPerSector       : Word;    (* Bytes per sector                                             *)
                     _HighestSector        : Byte;    (* Highest sector number within a cluster                       *)
                     _ShiftCount           : Byte;    (* Shift count to convert clusters into sectors                 *)
                     _ReservedSectors      : Word;    (* Number of reserved sectors at beginning of drive             *)
                     _FAT                  : Byte;    (* Number of FATs                                               *)
                     _RootEntries          : Word;    (* Number of root directory entries                             *)
                     _FirstDataSector      : Word;    (* Number of first sector containing user data                  *)
                     _HighestCluster       : Word;    (* Highest cluster number (number of data clsters + 1)
                                                         16-bit FAT if greater than 0FF6h, else 12-bit FAT            *)
                     _SectorsPerFAT        : Byte;    (* Number of sectors per FAT                                    *)
                     _FirstDirectorySector : Word;    (* Sector number of first directory sector                      *)
                     _DeviceDriverHeader   : Pointer; (* Address of device driver header                              *)
                     _MediaID              : Byte;    (* Media ID Byte                                                *)
                     _DiskAccessed         : Byte;    (* 00h if disk accessed, FFh if not                             *)
                     _NextDPB              : PDPB;    (* Pointer to next DPB                                          *)
                     _FreeCluster          : Word;    (* Cluster at which to start search for free space when writing *)
                     _FreeClustersNumber   : Word;);  (* Number of free clusters on drive, FFFFh = unknown            *)

                    (* For DOS 4.0+ *)
                1 : (Drive                : Byte;     (* 00h = A:, 01h = B:, etc                                      *)
                     UnitNumber           : Byte;     (* Unit number within device driver                             *)
                     BytesPerSector       : Word;     (* Bytes per sector                                             *)
                     HighestSector        : Byte;     (* Highest sector number within a cluster                       *)
                     ShiftCount           : Byte;     (* Shift count to convert clusters into sectors                 *)
                     ReservedSectors      : Word;     (* Number of reserved sectors at beginning of drive             *)
                     FAT                  : Byte;     (* Number of FATs                                               *)
                     RootEntries          : Word;     (* Number of root directory entries                             *)
                     FirstDataSector      : Word;     (* Number of first sector containing user data                  *)
                     HighestCluster       : Word;     (* Highest cluster number (number of data clsters + 1)
                                                         16-bit FAT if greater than 0FF6h, else 12-bit FAT            *)
                     SectorsPerFAT        : Word;     (* Number of sectors per FAT                                    *)
                     FirstDirectorySector : Word;     (* Sector number of first directory sector                      *)
                     DeviceDriverHeader   : Pointer;  (* Address of device driver header                              *)
                     MediaID              : Byte;     (* Media ID Byte                                                *)
                     DiskAccessed         : Byte;     (* 00h if disk accessed, FFh if not                             *)
                     NextDPB              : PDPB;     (* Pointer to next DPB                                          *)
                     FreeCluster          : Word;     (* Cluster at which to start search for free space when writing,
                                                         usually of free clusters allocated                           *)
                     FreeClustersNumber   : Word;);   (* Number of free clusters on drive, FFFFh = unknown            *)
                   end;

  PIOCTL = ^TIOCTL;
  TIOCTL = record
             SpecialFunctions  : Byte;  (* bit 0 set if function to use current BPB, clear if Device
                                                 BIOS Parameter Block field contains new default BPB
                                           bit 1 set if function to use track layout fields only
                                                 must be clear if CL=60h
                                           bit 2 set if all sectors in track same size (should be set)
                                           bits 3-7 reserved *)
             DeviceType        : Byte;  (* 00h    320K/360K disk
                                           01h    1.2M disk
                                           02h    720K disk
                                           03h    single-density 8-inch disk
                                           04h    double-density 8-inch disk
                                           05h    fixed disk
                                           06h    tape drive
                                           07h    (DOS 3.3+) other type of block device, normally 1.44M floppy
                                           08h    read/write optical disk
                                           09h    (DOS 5+) 2.88M floppy *)
             DeviceAttributes  : Word;  (* bit 0 set if nonremovable medium
                                           bit 1 set if door lock ("changeline") supported
                                           bits 2-15 reserved  *)
             NumberOfCylinders : Word;  (* number of cylinders *)
             MediaType         : Byte;  (* for 1.2M drive
                                           00h 1.2M disk (default)
                                           01h 320K/360K disk
                                           F8h for DUBLDISK.SYS v2.6 expanded drives
                                           always 00h for other drive types *)
             DeviceBPB         : Array [1..31] of Byte;  (* device BPB (see AH=53h), bytes after BPB offset 1Eh omitted *)
(* ---function 40h only--- *)
             SectorsPerTrack   : Word;  (* number of sectors per track (start of track layout field)
                                           max 63 *)
             SectorSizeInTrack : Word;  (* N Word pairs: number,size of each sector in track *)
           end;

Var
  DPB   : PDPB;
  IOCTL : PIOCTL;

{------------------------------------------------------------------------------}

function Bit (var P; D : Byte) : Boolean;

var
  L : Longint absolute P;

begin (* Bit *)
  Bit:=Odd (L shr D);
end;  (* Bit *)

{------------------------------------------------------------------------------}

function GetDriveIOCTL (var IOCTL : TIOCTL; Drive : Byte) : Boolean;

var
  Regs        : {$IFDEF MSDOS}
                  Registers;
                {$ELSE}
                  TRegisters;
                {$ENDIF}
  DOSPointer  : Pointer;
  DPMIPointer : Pointer;

begin  (* GetDriveIOCTL *)
  GetDriveIOCTL:=False;
  FillChar (IOCTL, SizeOf (IOCTL),  #0);
  If not GetDOSMem (DOSPointer, DPMIPointer, SizeOf (IOCTL)) then Exit;
  Move (IOCTL, DPMIPointer^, SizeOf (IOCTL));
  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$440D;
  Regs.BL:=Drive;
  Regs.CH:=$08;
  Regs.CL:=$60;
  Regs.DS:=PtrRec (DOSPointer).Seg;
  Regs.DX:=PtrRec (DOSPointer).Ofs;
  If not RealIntr ($21, Regs) or (Regs.Flags and FCarry = FCarry) then
    begin
      FreeDOSMem (DPMIPointer, SizeOf (IOCTL));
      Exit;
    end;
  Move (DPMIPointer^, IOCTL, SizeOf (IOCTL));
  FreeDOSMem (DPMIPointer, SizeOf (IOCTL));
  GetDriveIOCTL:=True;
end;   (* GetDriveIOCTL *)

{------------------------------------------------------------------------------}

var
  StackedDrives : Set of Byte;

function CompressedDrive (Drive : Byte) : TDriveMedia;

var
  Regs : {$IFDEF MSDOS}
           Registers;
         {$ELSE}
           TRegisters;
         {$ENDIF}

  {............................................................................}

  function Stacker : Boolean;

  type
    PStackerDevice = ^TStackerDevice;
    TStackerDevice = record
                       Signature    : Word;  (* A55Ah *)
                       Version      : Word;
                       OffsetTable  : Word;
                       Unknown1     : Array [1..56] of Byte;
                       VolumeNumber : Byte;
                       Unknown2     : Array [1..19] of Byte;
                       Swap         : Array [1..4] of Char;  (* 'SWAP' *)
                       Mapping      : TDrivesArray;
                   end;

  var
    StackerDevice        : PStackerDevice;
    StackerDeviceAddress : Pointer;
    DOSPointer           : Pointer;

  begin  (* Stacker *)
    Stacker:=False;

    If not GetDOSMem (DOSPointer, StackerDeviceAddress, SizeOf (DOSPointer)) then Exit;
    FillChar (Regs, SizeOf (Regs), #0);
    Regs.AX:=$4404;
    Regs.BL:=Drive;
    Regs.CX:=4;
    Regs.DS:=PtrRec (DOSPointer).Seg;
    Regs.DX:=PtrRec (DOSPointer).Ofs;
    If not RealIntr ($21, Regs) then
      begin
        FreeDOSMem (StackerDeviceAddress, SizeOf (DOSPointer));
        Exit;
      end;

    DOSPointer:=Pointer (StackerDeviceAddress^);
    FreeDOSMem (StackerDeviceAddress, SizeOf (DOSPointer));

    If not GetMappedDPMIPointer (StackerDevice, DOSPointer) then Exit;

    If StackerDevice^.Signature = $A55A then
      begin
        If StackerDevice^.Mapping [Drive] in StackedDrives then
          begin
            FreeMappedDPMIPointer (StackerDevice);
            Exit;
          end;
        StackedDrives:=StackedDrives + [Drive - 1];
        Stacker:=True;
      end;
    FreeMappedDPMIPointer (StackerDevice);
  end;   (* Stacker *)

  {............................................................................}

begin  (* CompressedDrive *)
  CompressedDrive:=INVALID_DRIVE;
  If IOCTL^.MediaType = $F8 then
    begin
      FillChar (Regs, SizeOf (Regs), #0);
      Regs.AX:=$4A11;
      Regs.BX:=$0001;
      Regs.DL:=Drive - 1;
      If not RealIntr ($2F, Regs) then Exit;
      If Stacker then
        begin
          CompressedDrive:=STACKER_DRIVE;
          Exit;
        end;
      If (Regs.AX = $0000) and (Bit (Regs.BL, 7)) then CompressedDrive:=DOUBLESPACE_DRIVE
      Else CompressedDrive:=DOUBLEDISK_DRIVE;
    end;
end;   (* CompressedDrive *)

{------------------------------------------------------------------------------}

function DriveIsCDROM (Drive : Byte) : Boolean;

var
  Regs : {$IFDEF MSDOS}
           Registers;
         {$ELSE}
           TRegisters;
         {$ENDIF}

begin (* DriveIsCDROM *)
  DriveIsCDROM:=False;

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$1500;
  If not RealIntr ($2F, Regs) then Exit;
  If Regs.BX = 0 then Exit;    (* CDROM extensions not installed *)

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$150C;
  If not RealIntr ($2F, Regs) then Exit;
  If Regs.BH < 2 then Exit;    (* CDROM extensions must be >= 2.00 *)

  Dec (Drive);

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$150B;
  Regs.CX:=Drive;
  If not RealIntr ($2F, Regs) then Exit;
  DriveIsCDROM:=(Regs.BX = $ADAD) and (Regs.AX <> $0000);
end;  (* DriveIsCDROM *)

{------------------------------------------------------------------------------}

function DriveIsFloppy (Drive : Byte) : Boolean;

begin  (* DriveIsFloppy *)
  DriveIsFloppy:=False;
  DriveIsFloppy:=IOCTL^.DeviceType in [00..04, 07, 09];
end;  (* DriveIsFloppy *)

{------------------------------------------------------------------------------}

function DriveIsOptical (Drive : Byte) : Boolean;

begin  (* DriveIsOptical *)
  DriveIsOptical:=False;
  DriveIsOptical:=IOCTL^.DeviceType = 08;
end;  (* DriveIsOptical *)

{------------------------------------------------------------------------------}

function DriveIsTape (Drive : Byte) : Boolean;

begin  (* DriveIsTape *)
  DriveIsTape:=False;
  DriveIsTape:=IOCTL^.DeviceType = 06;
end;  (* DriveIsTape *)

{------------------------------------------------------------------------------}

function DriveIsSubsted (Drive : Byte) : Boolean;

var
  Regs : {$IFDEF MSDOS}
           Registers;
         {$ELSE}
           TRegisters;
         {$ENDIF}

begin  (* DriveIsSubsted *)
  DriveIsSubsted:=False;
  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$4409;
  Regs.BL:=Drive;
  If not RealIntr ($21, Regs) then Exit;
  If FCarry and Regs.Flags <> 0 then Exit;
  DriveIsSubsted:=Bit (Regs.DX, 15);
end;  (* DriveIsSubsted *)

{------------------------------------------------------------------------------}

function DriveIsAssigned (Drive : Byte) : Byte;

var
  Regs        : {$IFDEF MSDOS}
                 Registers;
               {$ELSE}
                 TRegisters;
               {$ENDIF}
  Assign      : TDrivesArray;
  DPMIPointer : Pointer;

begin  (* DriveIsAssigned *)
  DriveIsAssigned:=Drive;

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$600;
  If not RealIntr ($2F, Regs) then Exit;
  If Regs.AL <> $FF then Exit;

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AX:=$601;
  If not RealIntr ($2F, Regs) then Exit;
  If Regs.Flags and FCarry <> 0 then Exit;

  If not GetMappedDPMIPointer (DPMIPointer, Ptr (Regs.ES, $103)) then Exit;

  Move (DPMIPointer^, Assign [1], SizeOf (Assign));
  FreeMappedDPMIPointer (DPMIPointer);
  DriveIsAssigned:=Assign [Drive];
end;   (* DriveIsAssigned *)

{------------------------------------------------------------------------------}

{$IFNDEF WINDOWS}
  function GetDriveType (Drive : Byte) : Word;

  var
    Regs : {$IFDEF MSDOS}
             Registers;
           {$ELSE}
             TRegisters;
           {$ENDIF}

  begin  (* GetDriveType *)
    GetDriveType:=0;
    FillChar (Regs, SizeOf (Regs), #0);
    Regs.AX:=$4408;
    Regs.BL:=Drive + 1;
    If not RealIntr ($21, Regs) then Exit;
    If Regs.Flags and FCarry = FCarry then
      begin
        FillChar (Regs, SizeOf (Regs), #0);
        Regs.AX:=$4409;
        Regs.BL:=Drive + 1;
        If not RealIntr ($21, Regs) then Exit;
        If Regs.Flags and FCarry = FCarry then Exit;
        If Bit (Regs.DX, 12) then GetDriveType:=DRIVE_REMOTE;
        Exit;
      end;
    If Regs.AX = 0 then GetDriveType:=DRIVE_REMOVABLE;
    If Regs.AX = 1 then GetDriveType:=DRIVE_FIXED;
  end;   (* GetDriveType *)
{$ENDIF}

{------------------------------------------------------------------------------}

procedure ConvertDPB (P : P1stDPB; var DPB : PDPB);

begin  (* ConvertDPB *)
  DPB^.Drive          :=P^.Drive;
  DPB^.UnitNumber     :=P^.UnitNumber;
  DPB^.BytesPerSector :=P^.BytesPerSector;
  DPB^.HighestSector  :=P^.HighestSector;
  DPB^.ShiftCount     :=P^.ShiftCount;
  DPB^.ReservedSectors:=P^.ReservedSectors;
  DPB^.FAT            :=P^.FAT;
  DPB^.RootEntries    :=P^.RootEntries;
  DPB^.FirstDataSector:=P^.FirstDataSector;
  DPB^.HighestCluster :=P^.HighestCluster;
  If Lo (DosVersion) < 4 then
    begin
      DPB^.SectorsPerFAT       :=P^._SectorsPerFAT;
      DPB^.FirstDirectorySector:=P^._FirstDirectorySector;
      DPB^.DeviceDriverHeader  :=P^._DeviceDriverHeader;
      DPB^.MediaID             :=P^._MediaID;
      DPB^.DiskAccessed        :=P^._DiskAccessed;
      DPB^.NextDPB             :=P^._NextDPB;
      DPB^.FreeCluster         :=P^._FreeCluster;
      DPB^.FreeClustersNumber  :=P^._FreeClustersNumber;
    end
  Else
    begin
      DPB^.SectorsPerFAT       :=P^.SectorsPerFAT;
      DPB^.FirstDirectorySector:=P^.FirstDirectorySector;
      DPB^.DeviceDriverHeader  :=P^.DeviceDriverHeader;
      DPB^.MediaID             :=P^.MediaID;
      DPB^.DiskAccessed        :=P^.DiskAccessed;
      DPB^.NextDPB             :=P^.NextDPB;
      DPB^.FreeCluster         :=P^.FreeCluster;
      DPB^.FreeClustersNumber  :=P^.FreeClustersNumber;
    end;
  DPB^.FAT16:=DPB^.HighestCluster > $FF6;
  DPB^.EntriesPerSector:=DPB^.BytesPerSector div 32;
  DPB^.MaxSectors      :=DPB^.RootEntries div DPB^.EntriesPerSector;
end;   (* ConvertDPB *)

{------------------------------------------------------------------------------}

function GetDPB (Drive : Byte) : Boolean;

var
  Regs : {$IFDEF MSDOS}
           Registers;
         {$ELSE}
           TRegisters;
         {$ENDIF}
  P    : P1stDPB;

begin  (* GetDPB *)
  GetDPB:=False;

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AH:=$32;
  Regs.DL:=Drive;
  If not RealIntr ($21, Regs) then Exit;
  If Regs.AL = $FF then Exit;

  If not GetMappedDPMIPointer (P, Ptr (Regs.DS, Regs.BX)) then Exit;

  ConvertDPB (P, DPB);
  FreeMappedDPMIPointer (P);
  GetDPB:=True;
end;   (* GetDPB *)

{------------------------------------------------------------------------------}

function DriveType (Drive : Byte) : TDriveMedia;

  {............................................................................}

  procedure AskInt21 (IOCTL : Boolean);

  var
    DType : TDriveMedia;

  begin   (* AskInt21 *)
    Case GetDriveType (Drive - 1) of
      DRIVE_REMOVABLE : begin
                          If GetDPB (Drive) then
                            begin
                              If DPB^.FAT = 1 then DriveType:=RAM_DRIVE
                              Else If (DPB^.MediaID = $FD) and
                                      (DPB^.SectorsPerFAT <> 2) then DriveType:=BERNOULLI_DRIVE
                                   Else DriveType:=REMOVABLE_DRIVE;
                            end
                          Else DriveType:=REMOVABLE_DRIVE;
                        end;

      DRIVE_FIXED : begin
                      If IOCTL then
                        begin
                          DType:=CompressedDrive (Drive);
                          If DType <> INVALID_DRIVE then
                            begin
                              DriveType:=DType;
                              Exit;
                            end;
                        end;
                      If GetDPB (Drive) then
                        begin
                          If DPB^.FAT = 1 then DriveType:=RAM_DRIVE
                          Else If (DPB^.MediaID = $FD) and
                                  (DPB^.SectorsPerFAT <> 2) then DriveType:=BERNOULLI_DRIVE
                               Else DriveType:=FIXED_DRIVE;
                        end
                      Else  DriveType:=FIXED_DRIVE;
                    end;

      DRIVE_REMOTE : DriveType:=LAN_DRIVE;

      Else DriveType:=INVALID_DRIVE;
    end;
  end;   (* AskInt21 *)

  {............................................................................}

begin (* DriveType *)
  If DriveIsSubsted (Drive) then DriveType:=SUBSTED_DRIVE
  Else If DriveIsAssigned (Drive) <> Drive then DriveType:=ASSIGNED_DRIVE
       Else If DriveIsCDROM (Drive) then DriveType:=CDROM_DRIVE
            Else If GetDriveIOCTL (IOCTL^, Drive) then
                   begin
                     If DriveIsFloppy (Drive) then DriveType:=FLOPPY_DRIVE
                     Else If DriveIsTape (Drive) then DriveType:=TAPE_DRIVE
                          Else If DriveIsOptical (Drive) then DriveType:=OPTICAL_DRIVE
                               Else AskInt21 (True);
                   end
                 Else AskInt21 (False);
end;  (* DriveType *)

{------------------------------------------------------------------------------}

function TruePath (Path : String) : String;

var
  STW          : Array [0..255] of Char;
  Regs         : {$IFDEF MSDOS}
                   Registers;
                 {$ELSE}
                   TRegisters;
                 {$ENDIF}
  DOSPointer1  : Pointer;
  DPMIPointer1 : Pointer;
  DOSPointer2  : Pointer;
  DPMIPointer2 : Pointer;

begin  (* TruePath *)
  TruePath:=Path;
  StrPCopy (STW, Path);

  If not GetDOSMem (DOSPointer1, DPMIPointer1, SizeOf (STW)) then Exit;
  Move (STW, DPMIPointer1^, SizeOf (STW));

  If not GetDOSMem (DOSPointer2, DPMIPointer2, SizeOf (STW)) then
    begin
      FreeDOSMem (DPMIPointer2, SizeOf (STW));
      Exit;
    end;

  FillChar (Regs, SizeOf (Regs), #0);
  Regs.AH:=$60;
  Regs.DS:=PtrRec (DOSPointer1).Seg;
  Regs.SI:=PtrRec (DOSPointer1).Ofs;
  Regs.ES:=PtrRec (DOSPointer2).Seg;
  Regs.DI:=PtrRec (DOSPointer2).Ofs;
  If (not RealIntr ($21, Regs)) or (Regs.Flags and FCarry = FCarry) then
    begin
      FreeDOSMem (DPMIPointer2, SizeOf (STW));
      FreeDOSMem (DPMIPointer1, SizeOf (STW));
      Exit;
    end;

  Move (DPMIPointer2^, STW, SizeOf (STW));
  TruePath:=StrPas (STW);
  FreeDOSMem (DPMIPointer2, SizeOf (STW));
  FreeDOSMem (DPMIPointer1, SizeOf (STW));
end;   (* TruePath *)

{----------------------------------------------------------------------------}

procedure FindDrives;

var
  I : Byte;

begin  (* FindDrives *)
  FillChar (Drives^, SizeOf (Drives^), #0);
  StackedDrives:=[];
  For I:=1 to 26 do
    begin
      Drives^[I].Media:=DriveType (I);

      Case Drives^[I].Media of
        INVALID_DRIVE : Continue;

        SUBSTED_DRIVE : Drives^[I].Path:=TruePath (Chr (I + 64) + ':\.');

        ASSIGNED_DRIVE : Drives^[I].Path:=Chr (DriveIsAssigned (I) + 64) + ':';

        Else Drives^[I].Path:=Char (I + 64) + ':';
      end;
    end;
end;   (* FindDrives *)

{------------------------------------------------------------------------------}

function Media2String;

begin  (* Media2String *)
  Case Media of
    FLOPPY_DRIVE      : Media2String:='Floppy Drive';
    REMOVABLE_DRIVE   : Media2String:='Removable Drive';
    FIXED_DRIVE       : Media2String:='Fixed Drive';
    LAN_DRIVE         : Media2String:='LAN Drive';
    RAM_DRIVE         : Media2String:='RAM Drive';
    CDROM_DRIVE       : Media2String:='CD-ROM Drive';
    BERNOULLI_DRIVE   : Media2String:='Bernoulli Drive';
    SUBSTED_DRIVE     : Media2String:='Substed Drive';
    ASSIGNED_DRIVE    : Media2String:='Assigned Drive';
    DOUBLEDISK_DRIVE  : Media2String:='Doubledisk Drive';
    DOUBLESPACE_DRIVE : Media2String:='Doublespace Drive';
    STACKER_DRIVE     : Media2String:='Stacker Drive';
    TAPE_DRIVE        : Media2String:='Tape Drive';
    OPTICAL_DRIVE     : Media2String:='Optical Drive';
    Else                Media2String:='Invalid Drive.';
  end;
end;   (* Media2String *)

{------------------------------------------------------------------------------}

var
  ExitSave : Pointer;

procedure MyExitProc;

begin  (* MyExitProc *)
  ExitProc:=ExitSave;
  Dispose (Drives);
  Dispose (IOCTL);
  Dispose (DPB);
end;   (* MyExitProc *)

{------------------------------------------------------------------------------}

Begin
  New (DPB);
  New (IOCTL);
  New (Drives);
  FillChar (Drives^, SizeOf (Drives^), #0);
  ExitSave:=ExitProc;
  ExitProc:=@MyExitProc;
End.
